import logging


class NetworkPlane_Checker():

    def __init__(self, address):
        self.__address = address

    def isString(self, x):
        return isinstance(x, str)

    def isInt(self, x):
        return isinstance(x, int)

    def isBoolean(self, x):
        return isinstance(x, bool)

    def compression_Algo(self, algo):

        if algo == "snappy" or algo == "":
            return True

        else:
            logging.getLogger(self.__address).error("Compression Error - Compression algo: " + str(
                algo) + " is not supported, only supporting: snappy")
            return False

    def isTypeValid(self, type):
        if type.lower()=="async":
            return True
        else:
            return False


class ConnectionError(Exception):

   def __init__(self, message):
      super().__init__(message)
      self.message = message

   def __str__(self):
      return self.message

class NetworkPlaneFailure(Exception):

   def __init__(self, message):
      super().__init__(message)
      self.message = message

   def __str__(self):
      return self.message

class InterfaceAttachFailed(Exception):

   def __init__(self, message):
      super().__init__(message)
      self.message = message

   def __str__(self):
      return self.message

class InterfaceRegistrationFailed(Exception):

   def __init__(self, message):
      super().__init__(message)
      self.message = message

   def __str__(self):
      return self.message



class UpdateRejected(Exception):

   def __init__(self, message):
      super().__init__(message)
      self.message = message

   def __str__(self):
      return self.message

class UpdateFailed(Exception):

   def __init__(self, message):
      super().__init__(message)
      self.message = message

   def __str__(self):
      return self.message


from cube_plugin_async_w_kafka.Networkplane.Network_Plane import Network_Plane
from cube_plugin_async_w_kafka.Networkplane.Exceptions import ConnectionError, InterfaceAttachFailed, NetworkPlaneFailure
from confluent_kafka import Producer

import queue
import logging

from multiprocessing import Process, Queue



"""
This network plane is assynchronious non transactional, it can send records with different TOPICS/KEYS while ensuring no duplication of messages and order delivery even during network failure
"""
class A_Network_Plane(Network_Plane):


    def __init__(self,network_plane_spec,to_be_connected_interfaces_spec):
        Network_Plane.__init__(self,network_plane_spec,to_be_connected_interfaces_spec)



    def start(self, timeout=10):
        """
        Starts the Network plane

        :param timeout:

            because of network jitters and communication quality:
                if timeout >= 0, NP will wait for that time (maximum) to discover topics and if it fails it will raise a NetworkPlaneFailure
                if timeout < 0, NP will wait indefinitely to discover all topics, it will start when discovery occurs

        :return:
        :raises: NetworkPlaneFailure
        """

        if not self._run_status.is_set():

            self._producer_config=self.get_producer_config(self._network_plane_spec)

            try:

                #wait to start for timeout, otherwise raise exception
                if timeout >= 0:

                    self.discover_topics(timeout)

                    self._run_status.set()
                    self._send_status.set()

                    self._burst_queue = Queue(self._network_plane_spec["buffering"]["burst-buffer-max-records"])
                    self._P_Live_Sync = Process(target=self._sync, args=(self._producer_config, self._burst_queue,self._address, self._send_status,self._task))
                    self._P_Live_Sync.start()

                    logging.getLogger(self._address).info("network plane is started")


                # wait to start when connection resumed, accept to receive messages, delivery will be done later
                else:

                    self._burst_queue = Queue(self._network_plane_spec["buffering"]["burst-buffer-max-records"])
                    self._P_Live_Sync = Process(target=self._sync, args=(self._producer_config, self._burst_queue,self._address, self._send_status,self._task))
                    self._P_Live_Sync.start()

                    logging.getLogger(self._address).warning(
                        "network plane will start automatically when connection with kafka cluster is up: "+self._producer_config['bootstrap.servers'])

                    #the system will block at this step until connection is established (topic are discovered)
                    self.discover_topics(-1)

                    self._run_status.set()
                    self._send_status.set()


            except ConnectionError as e:

                self._run_status.clear()
                self._send_status.clear()

                logging.getLogger(self._address).critical(
                        "network plane cannot start, no records will be accepted to be sent")

                raise NetworkPlaneFailure("network plane cannot start, no records will be accepted to be sent")



            except Exception as e:

                self._run_status.clear()
                self._send_status.clear()

                logging.getLogger(self._address).critical("unknown error, details: "+str(e))

                raise NetworkPlaneFailure("network plane cannot start, no records will be accepted to be sent")


        else:

            logging.getLogger(self._address).warning("network plane is already started or in starting process")


    def stop(self, timeout=None):
        """
        Stops the Network Plane
        :param timeout: time to wait to stop the sync thread
        :return:
        """

        if self._run_status.is_set():

            self._run_status.clear()

            if timeout is None:

                if not self._task.set("STOP", instructions={"timeout": None}):
                    pass


            else:
                if not self._task.set("STOP",instructions={"timeout": timeout}):
                    pass

            self._P_Live_Sync.join()

            self._send_status.clear()

            self._burst_queue.close()
            self._burst_queue.join_thread()

            logging.getLogger(self._address).info("network plane is closed with "+str(self._burst_queue.qsize())+" records in the queue")

            del self._P_Live_Sync

        else:
            logging.getLogger(self._address).warning("network plane is already closed or in closing process")


    def get_NetPort(self, interface_address):

        if interface_address in self._interfaces_status.keys():

            if self._interfaces_status[interface_address]["registered"] == True:

                self._interfaces_status[interface_address]["attached"] = True

                logging.getLogger(self._address).info(
                    "Attach succeeded: Interface " + str(interface_address) + " is attached to network plane " + str(
                        self._address))
                return self._burst_queue

            else:

                logging.getLogger(self._address).error(
                    "Attach Failed: Interface " + str(interface_address) + " is not registered, need to register first")
                raise InterfaceAttachFailed(
                    "Attach Failed: Interface " + str(interface_address) + " is not registered, need to register first")

        else:

            logging.getLogger(self._address).error(
                "Attach Failed: Interface " + str(interface_address) + " is not recognized by the network plane " + str(
                    self._address))
            raise InterfaceAttachFailed(
                "Attach Failed: Interface " + str(interface_address) + " is not recognized by the network plane " + str(
                    self._address))





    def _sync(self, __config, __burst_queue, __address, __send_status,__task):

        __producer = Producer(**__config, logger=logging.getLogger(__address))

        while True:

            # check if there is a task to do (update, close, ...)
            if not __task.to_do():

                try:

                    # wait for timeout, if no records after timeout: check for updates, otherwise: send that record and check next round
                    msg=__burst_queue.get(timeout=1000)

                    __producer.produce(topic=msg['topic'], key=msg['key'], value=msg['msg'], callback=self._on_delivery_callback)
                    __producer.poll(0)

                except queue.Empty:
                    pass

                #local queue is full : need to send messages in cache before start fetching new messages
                except BufferError as e:

                    __producer.poll(0)
                    __send_status.clear()
                    __timeout = 1

                    remaining_msg=__producer.flush(0)
                    b_percentage_msg_capacity=round((remaining_msg/__config['queue.buffering.max.messages'])*100,2)
                    b_kbytes_capacity=__config['queue.buffering.max.kbytes']

                    logging.getLogger(self._address).critical(
                        "Buffer bottleneck: " + str(
                        b_percentage_msg_capacity) + "% of its maximum record number ("+str(__config['queue.buffering.max.messages'])+"), maximum configured buffer size "+str(b_kbytes_capacity)+" kb)" + ", burst buffer contains: " + str(
                        __burst_queue.qsize()) + " records")

                    new_remaining_msg=remaining_msg

                    #loop until we start flushing
                    while new_remaining_msg >= remaining_msg:
                        new_remaining_msg=__producer.flush(__timeout)

                    __send_status.set()

                    b_percentage_msg_capacity = round(
                        (new_remaining_msg / __config['queue.buffering.max.messages']) * 100, 2)

                    logging.getLogger(self._address).critical(
                        "Buffer freed, network plane buffer capacity is " + str(
                            b_percentage_msg_capacity) + "% of its maximum record number, maximum configured buffer size is "+str(b_kbytes_capacity)+" kb" + ", burst buffer contains: " + str(
                            __burst_queue.qsize()) + " records")

                    #continue flashing the remaining records
                    __producer.flush()

                    #__producer.produce(topic=msg['topic'], key=msg['key'], value=msg['msg'], callback=self.on_delivery_callback)
                    #__producer.poll(0)

                except Exception as e:

                    __producer.poll(0)
                    __send_status.clear()

                    logging.getLogger(self._address).critical("Unknown delivery error: "+str(e))

            else:

                print("updating")

                task=__task.get_task()
                task_name=task["name"]
                task_instructions=task["instructions"]

                if task_name=="STOP":

                    logging.getLogger(self._address).warning("Interrupted for stopping task")
                    timeout=task_instructions["timeout"]

                    if timeout is not None:

                        if __producer.flush(timeout)==0:
                            __send_status.clear()
                            __task.success()

                            break
                        else:
                            __task.fail()

                    #wait until we flush all data
                    else:

                        __producer.flush()
                        __send_status.clear()
                        __task.success()

                elif task_name=="UPDATE_PRODUCER":

                    logging.getLogger(self._address).warning("Interrupted for producer update task")

                    #check the sending status prior to flush
                    if __send_status.is_set():

                        #flush remaining data
                        __producer.flush()

                        __saved_config=__config.copy()
                        for key, value in task_instructions.items():
                            __config[key]=value

                        try:

                            __producer = Producer(**__config, logger=logging.getLogger(__address))

                            __task.success()

                            del __saved_config

                            logging.getLogger(self._address).info(
                                        "Updates applied successfully, network plane updated with new instructions: " + str(
                                            task_instructions))

                                #config restore
                        except Exception as e:

                            logging.getLogger(self._address).error(
                                    "Updates application failed: producer encountered an error with the following configuration: "+str(__config)+", details: "+str(e))

                            __bad_config=__config
                            __config=__saved_config

                            __producer = Producer(**__config, logger=logging.getLogger(__address))

                            __task.fail()

                            logging.getLogger(self._address).info(
                                        "Undo updates: producer resumed successfully to last configuration: "+str(__config))

                            del __bad_config

                        #there is errors in sending, cannot update because cannot flush and create a new producer
                    else:

                            __task.fail()
                            logging.getLogger(self._address).error("Updates rejected: network plane is not working properly, need to establish a normal situation before proceeding with updates")

                else:

                    if task_name=="NONE":
                        __task.fail()
                        logging.getLogger(self._address).warning("Interruption ignored: no task mentioned")
                    else:
                        __task.fail()
                        logging.getLogger(self._address).warning("Interruption ignored: task "+task_name+" not identified")
















from SchemaV.Schema import SpecError
from cube_plugin_async_w_kafka.Networkplane.A_Network_Plane import A_Network_Plane

class NetManager():

    def __init__(self, configManager):


        self.__config_manager=configManager
        self.__address = configManager.getAppAddress()

        self.__isProvider_Validated=False

        self.__validated_network_planes={}

    def init(self, params):
        #check if provider config is OK
        pass

    def validate_provider(self, provider_config, params):
        """

        :param provider_config:
        :param params: params["verify_reachable"]: True/False
        :return:
        :raises: SpecError
        """

        self.__validated_provider=self.__config_manager.validate_provider(provider_config,params)
        self.__isProvider_Validated=True

        return self.__validated_provider

    def validate_network_plane(self,networkplane_config, params):
        """

        :param networkplane_config:
        :param params:
        :return:
        :raises: SpecError, AssertionError
        """

        if self.__isProvider_Validated:

            validated_networkplane= self.__config_manager.validate_networkplane(networkplane_config, params)
            self.__validated_network_planes[validated_networkplane["DCWN"]]={"config":validated_networkplane,"networkplane_object":None}

        else:

            try: dcwn=networkplane_config["DCWN"]
            except: raise SpecError("SPEC ERROR - Network plane should have a DCWN address")

            raise AssertionError("VALIDATION ERROR - Need to validate the provider specification related to network-plane DCWN: "+str(dcwn)+ " first.")

    def get_NetwokPlane(self, networkplane_dcwn, podds_dcwn_specs):
        """

        :param networkplane_dcwn:
        :param podds_dcwn_specs:
        :return:
        :raises: AssertionError
        """

        if networkplane_dcwn in self.__validated_network_planes.keys():

            networkplane=self.__validated_network_planes[networkplane_dcwn]

            if networkplane["networkplane_object"] is None:

                networkplane["networkplane_object"] = A_Network_Plane(networkplane["config"], podds_dcwn_specs)
                return networkplane["networkplane_object"]
            else:
                return networkplane["networkplane_object"]

        else:
            raise AssertionError("NETWORK PLANE BUILDING ERROR - Need to validate the network-plane DCWN: "+str(networkplane_dcwn)+" specification first.")














import time

from confluent_kafka.admin import AdminClient,ConfigResource,KafkaException
import confluent_kafka
import concurrent.futures


from cube_plugin_async_w_kafka.Networkplane.Exceptions import ConnectionError, InterfaceRegistrationFailed, InterfaceAttachFailed, NetworkPlaneFailure, UpdateFailed, UpdateRejected
import logging
from multiprocessing import Process, Event, Queue, Manager

from cube_plugin_async_w_kafka.Networkplane.Task import Task

class Network_Plane():
    """
    The main class to manage and live update kafka producers:
    - Regular: with/without data loss & duplication in case of failure.
    - Idempotent: no dataloss no duplication.
    - Transactional.

    Workflow:

            0 - The Network plane should start before starting the interfaces
            1 - It discovers the accessible topic (discover_topic method)
            2 - An interface should register to check if it's topic is known and accessible prior to get a NetPort
            3 - The interface sends records through the NetPort
    """

    def __init__(self,network_plane_spec, to_be_connected_interfaces_spec):

        self._network_plane_spec = network_plane_spec
        self._address = self._network_plane_spec["address"]

        #list of interfaces based on their address with their status: registered, attached ...
        self._interfaces_status = {}
        for interface_spec in to_be_connected_interfaces_spec:
            per_interface_status=Manager().dict()
            per_interface_status["interface_spec"]=interface_spec
            per_interface_status["registered"]=False
            per_interface_status["attached"]=False
            self._interfaces_status[interface_spec["address"]]=per_interface_status

        #gives indication: True/False if the network plane is running or stopped
        self._run_status = Event()
        self._run_status.clear()

        #gives indication: True/False if the network plane is sending data and operational
        self._send_status = Event()
        self._send_status.clear()

        #task handler
        self._task = Task()


        #topics seen by the network plane when connected to kafka broker
        self._accessible_topics = Manager().list()

        #topics that interfaces are working on
        self._active_topics=[]

        #the current producer config dict
        self._producer_config={}

        #stats
        self._latest_stats={}

        #Network plane status
        self.__status=Manager().dict()

        self._P_Live_Sync = None


    def isRunning(self):
        return self._run_status.is_set()

    def isConnected(self,connection_timeout=10):
        try:

            if connection_timeout > 0:

                admin_client = AdminClient(self._producer_config)
                topics=admin_client.list_topics(timeout=connection_timeout).topics
                logging.getLogger(self._address).critical("CHECK Connection: CONNECTION SUCCESS to kafka cluster " + str(self._producer_config['bootstrap.servers']))

                #considered as return True
                return topics

            else:

                logging.getLogger(self._address).error("CHECK Connection: TIMEOUT must be > 0")
                return False

        except Exception as e:

            logging.getLogger(self._address).critical("CHECK Connection: CONNECTION FAILED (timeout="+str(connection_timeout)+") to kafka cluster "+str(self._producer_config['bootstrap.servers'])+", details:"+str(e))
            return False


    def get_Status(self, connection_timeout=10):

        self.__status.clear()

        if connection_timeout > 0:

            self.__status["UP"] = self._run_status.is_set()
            self.__status["SEND"] = self._send_status.is_set()
            self.__status["RUNNING_UPDATE"] = self._task.to_do()
            self.__status["INT"] = self._interfaces_status
            self.__status["ACTIVE_TOPICS"] = self._active_topics
            self.__status["METRICS"]=self.metrics()



            #Check active topics
            discovered_topics=self.isConnected(connection_timeout)

            if discovered_topics:

                self.__status["CONNECTED"] = True
                self.__status["DISCOVERED_TOPICS"] = discovered_topics

                all_topics_still_active=True
                topic_name=""
                for topic in self._active_topics:

                    if not topic["name"] in discovered_topics.keys():
                        all_topics_still_active=False
                        topic_name=topic
                        break

                if all_topics_still_active:
                    logging.getLogger(self._address).info("Check TOPICS Operations - Topics SUCCESS - "+str(self._active_topics))
                    self.__status["ACTIVE_TOPICS_OPS"]=True
                else:
                    logging.getLogger(self._address).critical("Check TOPIC Operations - Topics FAILURE - "+str(topic_name)+" does not exist in the broker or not accessible")
                    self.__status["ACTIVE_TOPICS_OPS"] = False

            else:

                self.__status["CONNECTED"]=False
                self.__status["DISCOVERED_TOPICS"] = False
                self.__status["ACTIVE_TOPICS_OPS"]=False

        else:

            logging.getLogger(self._address).error("timeout must be > 0")

        return self.__status


    def update_producer(self, new_network_plane_spec,update_instructions, connection_timeout=10):
        """
        Live update of the producer config without traffic interruption.
        Only applicable when kafka cluster is reachable, it is not applicable when connection is lost (due to data in the producer instance that we cannot flush in this case).

        :param new_network_plane_spec:
        :param update_instructions:
        :return:
        :raises: UpdateFailed, UpdateRejected
        """


        if self.isConnected(connection_timeout) and self._run_status.is_set() :

            logging.getLogger(self._address).info("updating the network plane")


            if self._task.set("UPDATE_PRODUCER",update_instructions):

                self._network_plane_spec=new_network_plane_spec
                self._producer_config=self.get_producer_config(new_network_plane_spec)

            else:

                raise UpdateFailed("Update failed: task failed to execute")
        else:
            raise UpdateFailed("Update failed: network plane is not started or kafka cluster is unreachable")


    def discover_topics(self, connection_timeout=10):
        """
        Find the list of topic accessible by the network plane
        :param connection_timeout:
        :return:
        """

        try:

            admin_client = AdminClient(self._producer_config)
            topics_dict = admin_client.list_topics(timeout=connection_timeout).topics

            for topic, topicMetadata in topics_dict.items():

                if topic[:2] != '__':
                    self._accessible_topics.append(topic)


        except KafkaException as e:
            logging.getLogger(self._address).critical("kafka cluster unreachable:" +str(self._producer_config['bootstrap.servers']+", connection timed out: "+str(e)))
            raise ConnectionError("kafka cluster unreachable")


    def describe_topic(self, topic, connection_timeout=10):
        """
        Find the description of a specified topic
        :param topic:
        :param connection_timeout:
        :return:
        """

        if not topic in map(lambda x:x["name"],self._active_topics):

            working_topic={"name":topic}

            try:

                admin_client = AdminClient(self._producer_config)
                topic_dict = admin_client.list_topics(topic=topic,timeout=connection_timeout).topics
                topicMetadata=topic_dict[topic]

                topic_configResource = admin_client.describe_configs([ConfigResource(confluent_kafka.admin.RESOURCE_TOPIC, topic)])

                for j in concurrent.futures.as_completed(iter(topic_configResource.values())):
                    config_response = j.result(timeout=connection_timeout)

            except KafkaException as e:

                logging.getLogger(self._address).critical("kafka cluster unreachable:" + str(
                    self._producer_config['bootstrap.servers'] + ", connection timed out: " + str(e)))
                raise ConnectionError("kafka cluster unreachable")



            # data retention in hours
            working_topic["retention_hours"] = int(config_response['retention.ms'].value) / 3600000
            # minimum replica nodes which should send an ACK to say that data is sent
            working_topic["min_insync_replicas"] = int(config_response['min.insync.replicas'].value)

            # get partition metadata
            partitionsMetadata = topicMetadata.partitions


            # number of partitions inside the Topic
            working_topic["partitions"] = len(partitionsMetadata.keys())

            isTopicClean=True
            for part, partMetadata in partitionsMetadata.items():

                if partMetadata.error is not None:
                    isTopicClean=False
                    logging.getLogger(self._address).error("partition: "+str(part)+" in topic: "+topic+" has errors: "+str(partMetadata.error))
                #print(str(part) + ":" + str(partMetadata.replicas)  + ":" + str(partMetadata.error))

            working_topic["partitions_error"]= not isTopicClean

            self._active_topics.append(working_topic)
            return working_topic

        else:

            return list(filter(lambda x:x["name"]==topic,self._active_topics))[0]

    def register(self, interface_spec, connection_timeout=10):
        """
        Register an interface:
        - An interface is successfully registered if its topic matches one of the discovered topic by the network plane
        - If the topic is not discovered during network plane start (does not exist in KAFKA broker) the registration will fail and the interface will not be created)
        :param interface_spec:
        :param connection_timeout:
        :return:
        :raises: InterfaceRegistrationFailed
        """

        """ ---------------------------- Interface Registration ---------------------------- """


        interface_address=interface_spec['address']

        interface_topic=interface_spec['topic']

        try:self._interfaces_status[interface_address]
        except KeyError:
            raise InterfaceRegistrationFailed("registration rejected, interface: " + str(interface_address) + " is not recognized by network plane: " + str(self._address))



        if interface_topic in self._accessible_topics:

            if self._interfaces_status[interface_address]["registered"]==True:

                logging.getLogger(self._address).error("registration rejected, interface: " + str(interface_address) + " is already registered to network plane: " + str(self._address))

                raise InterfaceRegistrationFailed("registration rejected, interface: " + str(interface_address) + " is already registered to network plane: " + str(self._address))

            else:

                try:

                    topic_description=self.describe_topic(interface_topic, connection_timeout)
                    self._interfaces_status[interface_address]["registered"]=True

                    print("registered: "+interface_address)

                    logging.getLogger(self._address).info("interface: "+str(interface_address) + " is successfully registered to network plane: "+str(self._address)+", topic description: "+str(topic_description))

                    return topic_description

                except Exception as e:

                    logging.getLogger(self._address).error("registration failed: "+str(e))

                    raise InterfaceRegistrationFailed("registration failed: "+str(e))

        else:
            raise InterfaceRegistrationFailed("registration failed: interface with topic "+str(interface_topic)+" is not yet recognized by the network plane: "+str(self._address)+" or topic is not created in the kafka cluster")

    def reset(self,interface_address):
        self._interfaces_status[interface_address]["registered"]=False
        self._interfaces_status[interface_address]["attached"]=False




    def get_producer_config(self, network_plane_spec):
        """
        Builds kafka producer config from the network plane spec
        :param network_plane_spec:
        :return:
        """

        kconfig = {}

        # common configurations:
        kconfig['acks'] = -1
        kconfig['max.in.flight.requests.per.connection'] = 5
        kconfig['message.send.max.retries'] = 2147483647

        """ ---------------------------- ID ---------------------------- """
        # The client id is a user-specified string sent in each request to help trace calls.
        # It should logically identify the application and network plane making making the request including the name of cluster destination (to verify destination)
        kconfig['client.id'] = self._address
        # kconfig['client.rack']=network_plane_spec["region"]

        """ ---------------------------- BUFFERING ---------------------------- """

        # The maximum number of unsent messages that can be queued up the producer when using async mode before either the producer must be blocked or data must be dropped.
        kconfig['queue.buffering.max.messages'] = network_plane_spec["buffering"]["buffer-max-records"]

        # Maximum total message size sum allowed on the producer queue.
        # This queue is shared by all topics and partitions.
        # This property has higher priority than queue.buffering.max.messages.
        # default: 1048576 en Kb
        kconfig['queue.buffering.max.kbytes'] = network_plane_spec["buffering"]["buffer-max-size"]

        """ ---------------------------- BATCHING ---------------------------- """

        # Maximum number of messages batched in one MessageSet.
        # The total MessageSet size is also limited by batch.size and message.max.bytes.
        # default: 10000
        kconfig['batch.num.messages'] = network_plane_spec["batching"]["batch-max-records"]
        # Maximum size (in bytes) of all messages batched in one MessageSet, including protocol framing overhead.
        # This limit is applied after the first message has been added to the batch, regardless of the first message's size, this is to ensure that messages that exceed batch.size are produced.
        # The total MessageSet size is also limited by batch.num.messages and message.max.bytes.
        # debault: 1000000
        kconfig['batch.size'] = network_plane_spec["batching"]["batch-max-size"]

        # Delay in milliseconds to wait for messages in the producer queue to accumulate before constructing message batches (MessageSets) to transmit to brokers.
        # A higher value allows larger and more effective (less overhead, improved compression) batches of messages to accumulate at the expense of increased message delivery latency.
        # default 5
        # self._producer_config['queue.buffering.max.ms']=5
        kconfig['linger.ms'] = network_plane_spec["batching"]["batch-build-delay"]

        overhead = 512
        kconfig['message.max.bytes'] = network_plane_spec["batching"]["per-request-max-batches"] * \
                                       network_plane_spec["batching"]["batch-max-size"] + overhead

        """ ---------------------------- TIMEOUTS ---------------------------- """
        # Timeout for network requests (default: 60000 ms)
        kconfig['socket.timeout.ms'] = network_plane_spec["connection-timeout"]  # detect network failure after timeout

        # Local message timeout. This value is only enforced locally and limits the time a produced message waits for successful delivery.
        # A time of 0 is infinite. This is the maximum time librdkafka may use to deliver a message (including retries).
        # Delivery error occurs when either the retry count or the message timeout are exceeded.
        # The message timeout is automatically adjusted to transaction.timeout.ms if transactional.id is configured.
        kconfig['message.timeout.ms'] = network_plane_spec["on-failure"][
            "allow-data-loss"]  # if > 0, kafka will detect failed message after timeout and message failed to be sent will be dropped, a value of 0 will garanti that the message will not be dropped because failur because it will wait for ACK indefinitly

        """ ---------------------------- IDEMPOTENT PRODUCER ---------------------------- """

        if not network_plane_spec["on-failure"]["allow-data-duplication"]:
            kconfig['enable.idempotence'] = True
        else:
            kconfig['enable.idempotence'] = False

        """ ---------------------------- TRANSACTIONAL PRODUCER ---------------------------- """

        if network_plane_spec["type"] == "transactional":
            kconfig['transactional.id'] = self._address
            kconfig['transaction.timeout.ms'] = network_plane_spec["connection-timeout"]

        """ ---------------------------- DELIVERY REPORTS ---------------------------- """

        # Only provide delivery reports for failed messages.
        # default false
        # self._kconfig['delivery.report.only.error']=False

        """ ---------------------------- STATS COLLECTION ---------------------------- """

        if network_plane_spec["enable-stats"]:
            kconfig['statistics.interval.ms'] = 1000
            kconfig['stats_cb'] = self.__stats_callback
        else:
            kconfig['statistics.interval.ms'] = 0
            kconfig['stats_cb'] = self.__stats_callback

        """ ---------------------------- COMPRESSION ---------------------------- """
        kconfig['compression.codec'] = network_plane_spec["compression-codec"]

        """ ---------------------------- CONNECTION STRING ---------------------------- """
        connection_string = ""
        for broker in network_plane_spec["data-service"]["record-address"]:
            connection_string = connection_string + str(broker['host']) + ":" + str(broker['port']) + ","
        connection_string = connection_string[:-1]

        kconfig['bootstrap.servers'] = connection_string

        return kconfig

    def join(self):
        self._P_Live_Sync.join()

    def get_running_config(self):
        return {"producer":self._producer_config,"spec":self._network_plane_spec}

    def __stats_callback(self,stats):
        self._latest_stats=stats

    #will not be called if message.timeout.ms = 0 (kafka will wait indefinitly for message ack
    def _on_delivery_callback(self,err, msg):

        if err is not None:

            self._send_status.clear()
            logging.getLogger(self._address).error("delivery error: %s" % err)

        else:
            if not self._send_status.is_set():
                self._send_status.set()
                logging.getLogger(self._address).info("delivery resumed.")


    def metrics(self):
        stats=self._latest_stats
        return stats
    def _sync(self, __config, __burst_queue, __interruption_string, __address,__send_status,__working_status, __producer_update_task,__update_instructions):
        pass


import logging
import socket


class Provider_Checker():

    def __init__(self, address):
        self.__address = address

    def isString(self, x):
        return isinstance(x, str)

    def isInt(self, x):
        return isinstance(x, int)

    def isBoolean(self, x):
        return isinstance(x, bool)

    def isHostReachable(self, host):
        try:
            socket.gethostbyname(host)
            return True
        except socket.error:
            logging.getLogger(self.__address).error(
                "KAFKA BROKER UNREACHABLE - Host: " + str(host) + " is not reachable")
            return False

    def isValidPortNumber(self, port):
        if port >= 1 and port <= 65535:
            return True
        else:
            logging.getLogger(self.__address).error(
                "KAFKA BROKER PORT ADDRESS ERROR - Port: " + str(port) + " should be in range [1, 65535]")
            return False



from multiprocessing import Event, Manager
from ctypes import c_char_p


class Task():

    def __init__(self,):

        # Interruption: task to be done
        self.__to_do_task = Event()
        self.__to_do_task.clear()

        # Waiting event until finishing a task
        self.__finish_task = Event()
        self.__finish_task.set()

        # give an indication of a task status : success=True, failed=False
        self.__task_status = Event()
        self.__task_status.clear()

        self.__task_name=Manager().Value(c_char_p, "NONE")
        self.__instructions=Manager().dict()


    # Network plane method
    def set(self, task_name, instructions=None):

        #prepare
        self.__task_name.set(task_name)

        self.__instructions.clear()
        if isinstance(instructions, dict):
            for key, value in instructions.items():
                self.__instructions[key]=value

        self.__finish_task.clear()
        self.__task_status.clear()

        #interrupt & wait for the result
        self.__to_do_task.set()
        self.__finish_task.wait()

        #return the result after waiting for task execution
        return self.__task_status.is_set()




    # Producer method
    def success(self,):

        #clear interruption
        self.__to_do_task.clear()
        self.__task_name.set("NONE")

        self.__task_status.set()
        self.__finish_task.set()

    # Producer method
    def fail(self, ):

        self.__to_do_task.clear()
        self.__task_name.set("NONE")

        self.__task_status.clear()
        self.__finish_task.set()

    # Producer method
    def to_do(self):
        return self.__to_do_task.is_set()

    # Producer method
    def get_task(self):
        return {"name":self.__task_name.get(), "instructions":self.__instructions}








from SchemaV.Checker import Checker
from SchemaV.Schema import Validator

from cube_plugin_async_w_kafka.Checkers import NetworkPlane_Checker
from cube_plugin_async_w_kafka.Checkers import Provider_Checker
import json
import os

class ConfigManager():

    def __init__(self, address):

        self.__address = address
        self.__properties_file= os.path.join(os.path.dirname(os.path.abspath(__file__)),"properties.json")

        self.__provider_checker = Provider_Checker(address)
        self.__networkplane_checker = NetworkPlane_Checker(address)

    def load_properties(self):

        try:

            with open(self.__properties_file) as propf:
                properties = json.load(propf)
            propf.close()

        except Exception as e:
            raise ImportError("KAFKA OUTBOUND PROPERTIES FILE IMPORT ERROR - file :"+str(self.__properties_file)+" is not json consistent or does not exist, details: "+str(e))

        try:

            self.__version=properties["version"]

            self.__provider_name=properties["provider"]["name"]
            self.__check_reachable=properties["provider"]["check_reachable"]

            self.__io_policy=properties["networkplane"]["io_policy"]
            self.__type = properties["networkplane"]["type"]

            self.__default_compression_codec=properties["networkplane"]["compression_codec"]
            self.__default_enable_stats=properties["networkplane"]["enable_stats"]
            self.__default_connection_timeout=properties["networkplane"]["connection_timeout"]
            self.__default_allow_data_loss=properties["networkplane"]["allow_data_loss"]
            self.__default_allow_data_duplication=properties["networkplane"]["allow_data_duplication"]

            self.__default_buffer_max_records=properties["networkplane"]["buffer_max_records"]
            self.__default_buffer_max_size=properties["networkplane"]["buffer_max_size"]
            self.__default_burst_buffer_max_records=properties["networkplane"]["burst_buffer_max_records"]

            self.__default_batch_build_delay=properties["networkplane"]["batch_build_delay"]
            self.__default_batch_max_records=properties["networkplane"]["batch_max_records"]
            self.__default_batch_max_size=properties["networkplane"]["batch_max_size"]
            self.__default_per_request_max_batches=properties["networkplane"]["per_request_max_batches"]

        except Exception as e:
            raise ImportError("KAFKA OUTBOUND PROPERTIES FILE IMPORT ERROR - In file :"+str(self.__properties_file)+", details: "+str(e))

    def get_version(self):
        return self.__version

    def get_provider_schema(self, params):
        """

        :param params: params["check_reachable"]: True/False
        :return:
        """

        try:
            connectivity_check = params["check_reachable"]
        except:
            connectivity_check = self.__check_reachable

        if connectivity_check:
            host_checker = Checker(True, self.__provider_checker.isHostReachable, None)
        else:
            host_checker = Checker(True, None, None)

        schema = {'checkIn': None,
                  'checkNext': {
                      'name': {'checkIn': Checker(False, self.__provider_checker.isString, "NA"), 'checkNext': None},
                      'DCWN': {'checkIn': Checker(True, self.__provider_checker.isString, None), 'checkNext': None},
                      'region': {'checkIn': Checker(True, self.__provider_checker.isString, None), 'checkNext': None},
                      'record-address': {'checkIn': Checker(True, None, None),
                                         'checkNext': [{'checkIn': Checker(True, None, None), 'checkNext': {
                                             'host': {
                                                 'checkIn': host_checker,
                                                 'checkNext': None},
                                             'port': {
                                                 'checkIn': Checker(True, self.__provider_checker.isValidPortNumber,
                                                                    None),
                                                 'checkNext': None}}}
                                                       ]
                                         }
                  }

                  }

        return schema

    def get_networkplane_schema(self, params):

        return {'checkIn': None, 'checkNext': {

            'name': {'checkIn': Checker(False, self.__networkplane_checker.isString, "NA"), 'checkNext': None},
            'DCWN': {'checkIn': Checker(True, self.__networkplane_checker.isString, None), 'checkNext': None},
            'type': {'checkIn': Checker(True, self.__networkplane_checker.isTypeValid, None), 'checkNext': None},
            'data-service': {'checkIn': Checker(True, self.__networkplane_checker.isString, None), 'checkNext': None},
            'compression-codec': {'checkIn': Checker(False, self.__networkplane_checker.compression_Algo, self.__default_compression_codec),
                                  'checkNext': None},
            'enable-stats': {'checkIn': Checker(False, self.__networkplane_checker.isBoolean, self.__default_enable_stats), 'checkNext': None},
            'connection-timeout': {'checkIn': Checker(False, self.__networkplane_checker.isInt, self.__default_connection_timeout),
                                   'checkNext': None},

            'on-failure': {'checkIn': Checker(False, None, {'allow-data-loss': self.__default_allow_data_loss}), 'checkNext': {
                'allow-data-loss': {'checkIn': Checker(False, self.__networkplane_checker.isInt, self.__default_allow_data_loss), 'checkNext': None}}
                           },

            'on-recover': {'checkIn': Checker(False, None, {'allow-data-duplication': self.__default_allow_data_duplication}), 'checkNext': {
                'allow-data-duplication': {'checkIn': Checker(False, self.__networkplane_checker.isBoolean, self.__default_allow_data_duplication),
                                           'checkNext': None}}
                           },

            'buffering': {'checkIn': Checker(True, None, {'buffer-max-records': self.__default_buffer_max_records, 'buffer-max-size': self.__default_buffer_max_size,
                                                          "burst-buffer-max-records": self.__default_burst_buffer_max_records}), 'checkNext': {
                'buffer-max-records': {'checkIn': Checker(False, self.__networkplane_checker.isInt, self.__default_buffer_max_records),'checkNext': None},
                'buffer-max-size': {'checkIn': Checker(False, self.__networkplane_checker.isInt, self.__default_buffer_max_size),'checkNext': None},
                'burst-buffer-max-records': {'checkIn': Checker(False, self.__networkplane_checker.isInt, self.__default_burst_buffer_max_records),'checkNext': None}}
                          },

            'batching': {'checkIn': Checker(False, None, {'batch-build-delay': self.__default_batch_build_delay, 'batch-max-records': self.__default_batch_max_records,
                                                          "batch-max-size": self.__default_batch_max_size, 'per-request-max-batches': self.__default_per_request_max_batches}),
                         'checkNext': {
                             'batch-build-delay': {'checkIn': Checker(False, self.__networkplane_checker.isInt, self.__default_batch_build_delay),'checkNext': None},
                             'batch-max-records': {'checkIn': Checker(False, self.__networkplane_checker.isInt, self.__default_batch_max_records),'checkNext': None},
                             'batch-max-size': {'checkIn': Checker(False, self.__networkplane_checker.isInt, self.__default_batch_max_size),'checkNext': None},
                             'per-request-max-batches': {'checkIn': Checker(False, self.__networkplane_checker.isInt, self.__default_per_request_max_batches),'checkNext': None}}}
        }
                }

    def validate_provider(self,provider_config,params):
        """

        :param provider_config:
        :param params: params["verify_reachable"]: True/False
        :return:
        """

        provider_validator=Validator(self.get_provider_schema(params))
        return provider_validator.validate(provider_config)

    def validate_networkplane(self,networkplane_config, params):

        networkplane_validator = Validator(self.get_networkplane_schema(params))
        return networkplane_validator.validate(networkplane_config)



from setuptools import setup, find_packages

setup(

  name = 'cube_plugin_async_w_kafka',         # How you named your package folder (MyLib)
  packages = find_packages("."),
  include_package_data=True,
  version = '1.0.1',      # Start with a small number and increase it with every change you make
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'Kafka service provider for Outbound traffic in asynchronous mode',   # Give a short description about your library
  author = 'Abdallah Ben Hamida',                   # Type in your name
  author_email = 'noemail@gmail.com',      # Type in your E-Mail
  url = 'https://github.com/QuanticData',   # Provide either the link to your github or to your website
  #download_url = 'https://github.com/QuanticData/SchemaV/archive/refs/tags/v_01.tar.gz',    # I explain this later on
  keywords = ['dict', 'python', 'kafka'],   # Keywords that define your package best
  install_requires=[
      'confluent-kafka', #update this section with requirements file and include dependencies versions
      'schemav'

      ],
  classifiers=[
    'Development Status :: 5 - Production/Stable',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support

  ],
)

